# /bin/bash
import requests
from requests.exceptions import HTTPError
from trading212py.config import API_VERSION,ACCOUNT_TYPE,API_KEY
from typing import Literal, Optional,Dict,Any
from trading212py.base import (Position, AccountMetadata, AccountCash,
                               Exchange, Exchanges, Instrument, Instruments, Pie, PieListItem, PieList,
                               Order, HistoricalItem,HistoricalOrderResponseModel,
                               DividendResponseModel, ExportReport,ExportPayload,ExportReportResponse,
                               Transactions,TransactionPayload,CreatePie)
from trading212py.decorators import unpacker

class T212:
    def __init__(self) -> None:
        self._base_url: str = f'https://{ACCOUNT_TYPE}.trading212.com/api/{API_VERSION}'
        self._session: requests.Session = requests.Session()
        self._api_key: str | None = self._get_api_key()
    
    def _get_api_key(self) -> str:
        try:
            if API_KEY is None: raise Exception("T212_API_KEY environment variable couln't be retreived.")
            else: return API_KEY
        except KeyError as e:
            raise Exception(f"{e} - Set T212_API_KEY environment variable.")

    def __request(self, method:str=None, endpoint:str=None, query_params=None, json:Optional[Dict]=None, **kwargs) -> requests.Response:
        try:
            response: requests.Response = self._session.request(
                method=method.upper(),
                url=f'{self._base_url}{endpoint}',
                headers={
                    "Authorization": self._api_key,
                    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
                    "content-type": "application/json; charset=UTF-8",
                    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36",
                    "accept-encoding": "gzip, deflate, br",
                    "accept-language": "en-US,en;q=0.9"
                },
                json=json,
                params=query_params, **kwargs
            )
            response.raise_for_status()           
        except HTTPError as http_err:
            print(f'{http_err}')
        except Exception as e:
            if method == 'DELETE':
                if response.status_code == 404:
                    return response.text # Return None or a custom message indicating successful handling of 404
                else: print(f"Error making request: {response.text} - {e}")
            else: print(f"Error making request: {response.text} - {e}")
        else:
            return response.json() if len(response.json())!=0 else None


    # Account Data
    def _get_account_metadata(self):
        return self.__request(method='GET', endpoint='/equity/account/info')
    
    def _get_account_cash(self):
        return self.__request(method='GET', endpoint='/equity/account/cash')
    
    # Personal Portfolio
    def _get_portfolio(self):
        return self.__request(method='GET', endpoint='/equity/portfolio')
    
    def _get_portfolio_ticker(self, ticker):
        return self.__request(method='GET', endpoint=f'/equity/portfolio/{ticker}')

    # Instruments Metadata
    def _get_exchange_list(self):
        return self.__request(method='GET', endpoint='/equity/metadata/exchanges')
    
    def _get_instrument_list(self):
        return self.__request(method='GET', endpoint='/equity/metadata/instruments')
    
    # Pies
    def _get_pie_list(self):
        return self.__request(method='GET', endpoint='/equity/pies')
    
    def _post_create_pie(self, payload):
        return self.__request(method='POST', endpoint='/equity/pies', query_params=payload)
    
    def _delete_delete_pie(self, pie_id):
        return self.__request(method='DELETE', endpoint=f'/equity/pies/{pie_id}')

    def _get_pie(self, pie_id):
        return self.__request(method='GET', endpoint=f'/equity/pies/{pie_id}')
    
    def _post_update_pie(self, pie_id, payload):
        return self.__request(method='POST', endpoint=f'/equity/pies/{pie_id}', query_params=payload)

    # Equity Orders
    def _get_all_orders(self):
        return self.__request(method='GET', endpoint='/equity/orders')
    
    def _post_place_limit_order(self, payload):
        return self.__request(method='POST', endpoint='/equity/orders/limit', query_params=payload)
    
    def _post_place_market_order(self, payload):
        return self.__request(method='POST', endpoint='/equity/orders/market', query_params=payload)
    
    def _post_stop_order(self, payload):
        return self.__request(method='POST', endpoint='/equity/orders/stop', query_params=payload)
                              
    def _post_stop_limit_order(self, payload):
        return self.__request(method='POST', endpoint='/equity/orders/stop_limit', query_params=payload)
    
    def _delete_cancel_order(self, order_id) -> int:
        return self.__request(method='DELETE', endpoint=f'/equity/orders/{order_id}')
    
    def _get_order(self, order_id):
        return self.__request(method='GET', endpoint=f'/equity/orders/{order_id}')
    
    # Historical Items
    def _get_historical_orders(self,payload):
        return self.__request(method='GET', endpoint='/equity/history/orders', query_params=payload)
    
    def _get_dividends(self,payload): #TODO: #2 Dividends function is returning Response code 500 internal server error.
        return self.__request(method='GET', endpoint='/history/dividends', query_params=payload)
    
    def _get_exports_list(self):
        return self.__request(method='GET', endpoint='/history/exports')
    
    def _post_exports(self,payload):
        return self.__request(method='GET', endpoint=f'/history/exports', query_params=payload)
    
    def _get_transactions(self,payload):
        return self.__request(method='GET', endpoint='/history/transactions', query_params=payload)
    

    # Methods
    # Account Data
    @unpacker(cls=AccountMetadata)
    def account_metadata(self):
        '''Returns the account metadata, including account type, currency, and other account details.'''  
        return self._get_account_metadata()
    
    @unpacker(cls=AccountCash)
    def account_cash(self) -> list[AccountCash]:
        '''Returns the account cash balance.'''
        return self._get_account_cash()

    # Personal Portfolio
    @unpacker(cls=Position, clsList=True)
    def portfolio(self):
        '''Returns the portfolio, including the positions and metadata for each position.'''
        return self._get_portfolio()

    @unpacker(cls=Position)
    def portfolio_ticker(self, ticker) -> Position:
        '''Returns the portfolio details for a specific ticker.

        Args:
            ticker (str): The ticker symbol of the position to retrieve.
        '''
        return self._get_portfolio_ticker(ticker=ticker)

    # Instruments Metadata
    @unpacker(cls=Exchange, clsList=True)
    def exchange_list(self) -> Exchanges:
        '''Returns the list of exchanges supported by Trading212.'''
        return self._get_exchange_list()

    @unpacker(cls=Instrument, clsList=True)
    def instrument_list(self) -> Instruments:
        '''Returns the list of instruments supported by Trading212.'''
        return self._get_instrument_list()

    # Pies
    @unpacker(cls=PieListItem, clsList=True)
    def pie_list(self) -> PieList:
        '''Returns the list of Pies (Portfolio Investment Environments) by the user.
        API Ref: https://t212public-api-docs.redoc.ly/#operation/getAll
        '''
        return self._get_pie_list()

    @unpacker(cls=Pie, clsList=True)
    def create_pie(self, payload:CreatePie) -> Pie:
        '''Creates a new Pie (Portfolio Investment Environment) with the provided payload.
        Ref: https://t212public-api-docs.redoc.ly/#operation/create

        Args:
            payload (dict): The payload containing the details of the new Pie.
        '''
        return self._post_create_pie(payload=payload.model_dump(exclude_none=True))
    
    def delete_pie(self, pie_id:int=None) -> int:
        '''Deletes the Pie with the specified pie_id.

        Args:
            pie_id (str): The ID of the Pie to delete.
        '''
        if not pie_id: raise Exception(f"Provide {pie_id=}")
        else: return self._delete_delete_pie(pie_id=pie_id)

    @unpacker(cls=Pie)
    def pie(self, pie_id:int=None) -> Pie:
        '''Returns the details of the Pie with the specified pie_id.

        Args:
            pie_id (str): The ID of the Pie to retrieve.
        '''
        if not pie_id: raise Exception(f"Provide {pie_id=}")
        else: return self._get_pie(pie_id=pie_id)

    def update_pie(self, pie_id, payload):
        '''Updates the Pie with the specified pie_id using the provided payload.

        Args:
            pie_id (str): The ID of the Pie to update.
            payload (dict): The payload containing the updated details of the Pie.
        '''
        return self._post_update_pie(pie_id=pie_id, payload=payload)

    # Equity Orders
    @unpacker(cls=Order, clsList=True)
    def all_orders(self):
        '''Returns all orders placed by user.'''
        return self._get_all_orders()

    @unpacker(cls=Order)
    def place_limit_order(self, payload:Order):
        '''Places a limit order with the provided payload.

        Args:
            payload (dict): The payload containing the details of the limit order.
        '''
        return self._post_place_limit_order(payload=payload.model_dump(exclude_none=True))

    @unpacker(cls=Order)
    def place_market_order(self, payload:Order):
        '''Places a market order with the provided payload.

        Args:
            payload (dict): The payload containing the details of the market order.
        '''
        return self._post_place_market_order(payload=payload.model_dump(exclude_none=True))

    @unpacker(cls=Order)
    def stop_order(self, payload:Order):
        '''Places a stop order with the provided payload.

        Args:
            payload (dict): The payload containing the details of the stop order.
        '''
        return self._post_stop_order(payload=payload.model_dump(exclude_none=True))

    @unpacker(cls=Order)
    def stop_limit_order(self, payload:Order):
        '''Places a stop limit order with the provided payload.

        Args:
            payload: (dict) The payload containing the details of the stop limit order.
        '''
        return self._post_stop_limit_order(payload=payload.model_dump(exclude_none=True))

    def cancel_order(self, order_id:int=None) -> None:
        '''Cancels the order with the specified order_id.

        Args:
            order_id (str): The ID of the order to cancel.
        '''
        if not order_id: raise Exception(f"Provide {order_id=}")
        else: return self._delete_cancel_order(order_id=order_id)

    @unpacker(cls=Order)
    def order(self, order_id:int=None):
        '''Returns the details of the order with the specified order_id.

        Args:
            order_id (str): The ID of the order to retrieve.
        '''
        return self._get_order(order_id=order_id)

    # Historical Items
    @unpacker(cls=HistoricalOrderResponseModel)
    def historical_orders(self,payload:HistoricalItem):
        '''Returns the historical orders.      
        Args:
            HistoricalItem class
                cursor (int): The cursor value for pagination.
                limit (int): The maximum number of orders to retrieve.
                    Default: 20
                    Max items: 50
                    Example: limit=21
                ticker (str): The ticker symbol to filter orders by.

        Example:
                historicalItem = HistoricalItem(cursor=0,ticker='AAPL_US_EQ',limit=20)
                print(t212.historical_orders(payload=historicalItem))
        '''
        return self._get_historical_orders(payload=payload.model_dump(exclude_none=True))

    
    @unpacker(cls=DividendResponseModel)
    def dividends(self,payload:HistoricalItem):
        '''Returns the dividends.
        Args:
            HistoricalItem class
                cursor (int): The cursor value for pagination.
                limit (int): The maximum number of orders to retrieve.
                    Default: 20
                    Max items: 50
                    Example: limit=21
            ticker (str): The ticker symbol to filter orders by.
        '''
        return self._get_dividends(payload=payload.model_dump(exclude_none=True))

    @unpacker(cls=ExportReport)
    def exports_list(self):
        '''Returns the list of exports.
        NOTE: THIS ENDPOINT IS NOT AVAILABLE FOR IN DEMO MODE.
        '''
        if ACCOUNT_TYPE.upper()=='DEMO': raise Exception(f"This endpoint is not available in {ACCOUNT_TYPE=} mode.")
        else: return self._get_exports_list()

    @unpacker(cls=ExportReportResponse)
    def exports(self, payload:ExportPayload):
        '''Creates a new export with the provided payload.

        Args:
            payload (dict): The payload containing the details of the export.
        
        Example:
            from trading212py.base import ExportPayload
            payload = {
                "dataIncluded": {
                    "includeDividends": True,
                    "includeInterest": True,
                    "includeOrders": True,
                    "includeTransactions": True
                },
                "timeFrom": "2019-08-24T14:15:22Z",
                "timeTo": "2024-09-07T14:15:22Z"
                }
            print(t212.exports(payload=ExportPayload(**payload)))
        '''
        return self._post_exports(payload=payload.model_dump(exclude_none=True))

    @unpacker(cls=Transactions)
    def transactions(self,payload:TransactionPayload):
        '''Returns the transactions.
        
        Args:
            cursor (str) = Pagination cursor
            limit (int) = The maximum number of transactions to retrieve. Default: 20, Max items: 50
        '''
        return self._get_transactions(payload=payload.model_dump(exclude_none=True))


from functools import wraps
import json
from typing import Optional
# ...

def debug(func):
    """Print the function signature and return value"""
    @wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={repr(v)}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        print(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        print(f"{func.__name__}() returned {repr(value)}")
        return value
    return wrapper_debug

def unpacker(cls:object=None,clsList:Optional[bool]=False):
    '''Unpack items to a class
    '''
    def decorator(f):    
        @wraps(f)
        def wrapper(*args, **kwargs):
            # Do something before the function
            data = f(*args, **kwargs)
            # Do something after the function
            if data is None: return None
            else:
                if type(data) is not str:
                    if (cls is None and clsList is None) : return data
                    else:
                        if clsList is not False: return [cls(**item) for item in data]
                        else: return cls(**data)
                else: return data
        return wrapper
    return decorator

def jsondump(func):
    '''Dump the data as json
    '''
    @wraps(wrapped=func)
    def wrapper(*args, **kwargs):
        data = func(*args, **kwargs)
        out_file = open("jsondump1.json", "w")
        json.dump(obj=data,fp=out_file)
        out_file.close()
        return data
    return wrapper


import os
from dotenv import dotenv_values

# Set Trading212 API version
API_VERSION:str = 'v0'

InProduction = True

if os.getenv('IN_PRODUCTION', 'false') == 'false':
    # Load environment variables from .env/.secret or .env/.shared
    # Specify the path to the .env file if needed
    env_config = {
        **dotenv_values(dotenv_path=".env/.shared"),  # load shared development variables
        **dotenv_values(dotenv_path=".env/.secret"),  # load sensitive variables
        **os.environ,  # override loaded values with environment variables
    }
    # Set InProduction to False if not in production
    InProduction = False
else: pass

# Get the API key based on the account type
ACCOUNT_TYPE:str= env_config.get('ACCOUNT_TYPE', 'live')
API_KEY:str = env_config.get('T212_DEMO_API_KEY', None) if ACCOUNT_TYPE == 'demo' else env_config.get('T212_API_KEY', None)


from __future__ import annotations
from enum import Enum
from datetime import datetime
from typing import List, Optional, Dict,Any
from pydantic import BaseModel
from dataclasses import dataclass

class OrderStatus(Enum): #TODO: #1 Not yet implemented. 
    LOCAL = "LOCAL"
    UNCONFIRMED = "UNCONFIRMED"
    CONFIRMED = "CONFIRMED"
    NEW = "NEW"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"
    FILLED = "FILLED"
    REJECTED = "REJECTED"
    REPLACING = "REPLACING"
    REPLACED = "REPLACED"


class TransactionType(Enum):
    WITHDRAW = "WITHDRAW" 
    DEPOSIT = "DEPOSIT" 
    FEE = "FEE"
    TRANSFER = "TRANSFER"

'''
Account Data Classes
'''
class AccountMetadata(BaseModel):
    id:Optional[int]
    currencyCode:Optional[str]

class AccountCash(BaseModel):
    free:float
    total:float
    ppl:float
    result:float
    invested:float
    pieCash:float
    blocked:float | None

'''
Personal Portfolio Classes
'''
class Position(BaseModel):
    ticker: str
    quantity: float
    averagePrice: float
    currentPrice: float
    ppl: float
    fxPpl: float
    initialFillDate: str
    frontend: str
    maxBuy: float
    maxSell: float
    pieQuantity: float

# class Positions(BaseModel):
#     positions: List[Position]

'''
Instruments Metadata CLasses
'''
class TimeEvent(BaseModel):
    date: str
    type: str

class WorkingSchedule(BaseModel):
    id: int
    timeEvents: List[TimeEvent]

class Exchange(BaseModel):
    id: int
    name: str
    workingSchedules: Optional[List[WorkingSchedule]]

class Exchanges(BaseModel):
    exchanges: List[Exchange]


class Instrument(BaseModel):
    addedOn: str
    currencyCode:str
    isin:str
    maxOpenQuantity: int | float
    minTradeQuantity: int | float
    name:str
    shortname: Optional[str]
    ticker: str
    type: str
    workingScheduleId:int

class Instruments(BaseModel):
    instruments: List[Instrument]

'''
Pie Classes Section
'''
# Pie
class dividendCashAction(Enum):
    REINVESTED = "REINVESTED"
    TO_ACCOUNT_CASH = "TO_ACCOUNT_CASH"

class Icon(Enum):
    HOME = "Home"
    PIGGY_BANK = "PiggyBank"
    ICEBERG = "Iceberg"
    AIRPLANE = "Airplane"
    RV = "RV"
    UNICORN = "Unicorn"
    WHALE = "Whale"
    CONVERTIBLE = "Convertible"
    FAMILY = "Family"
    COINS = "Coins"
    EDUCATION = "Education"
    BILLS_AND_COINS = "BillsAndCoins"
    BILLS = "Bills"
    WATER = "Water"
    WIND = "Wind"
    CAR = "Car"
    BRIEFCASE = "Briefcase"
    MEDICAL = "Medical"
    LANDSCAPE = "Landscape"
    CHILD = "Child"
    VAULT = "Vault"
    TRAVEL = "Travel"
    CABIN = "Cabin"
    APARTMENTS = "Apartments"
    BURGER = "Burger"
    BUS = "Bus"
    ENERGY = "Energy"
    FACTORY = "Factory"
    GLOBAL = "Global"
    LEAF = "Leaf"
    MATERIALS = "Materials"
    PILL = "Pill"
    RING = "Ring"
    SHIPPING = "Shipping"
    STOREFRONT = "Storefront"
    TECH = "Tech"
    UMBRELLA = "Umbrella"
    NONE = ''

class PieIssueName(Enum):
    DELISTED = "DELISTED"
    SUSPENDED = "SUSPENDED"
    NO_LONGER_TRADABLE = "NO_LONGER_TRADABLE"
    MAX_POSITION_SIZE_REACHED = "MAX_POSITION_SIZE_REACHED"
    APPROACHING_MAX_POSITION_SIZE = "APPROACHING_MAX_POSITION_SIZE"
    COMPLEX_INSTRUMENT_APP_TEST_REQUIRED = "COMPLEX_INSTRUMENT_APP_TEST_REQUIRED"
    NONE = ''

class PieIssueTransactionType(Enum):
    IRREVERSIBLE = "IRREVERSIBLE"
    REVERSIBLE = "REVERSIBLE"
    INFORMATIVE = "INFORMATIVE"

class PieIssue(BaseModel):
    name: Optional[PieIssueName] = None
    severity: Optional[PieIssueTransactionType] = None

class Result(BaseModel):
    investedValue: Optional[float]= None
    result: Optional[float] = None
    resultCoef: Optional[float] = None
    value: Optional[float] = None

class PieInstrument(BaseModel):
    currentShare: Optional[float] = None
    expectedShare: Optional[float] = None
    issues: Optional[List[PieIssue]] = None
    ownedQuantity: Optional[float] = None
    result: Optional[Result] = None
    ticker: str = None

class PieSettings(BaseModel):
    creationDate: Optional[str] = None
    dividendCashAction: Optional[str] = None
    endDate: Optional[str] = None
    goal: Optional[float] = None
    icon: Optional[Icon] = None
    id: Optional[int] = None
    initialInvestment: Optional[float] = None
    instrumentShares: Optional[Dict[str, float]] = None
    name: Optional[str] = None
    pubicUrl:Optional[str] = None

class Pie(BaseModel):
    instruments: Optional[List[PieInstrument]] = None
    settings: Optional[PieSettings] = None

# Create pie
class CreatePie(BaseModel):
    dividendCashAction: str
    endDate: str
    goal: float
    icon: str
    instrumentShares: Dict[str, float]
    name: str

# Pie List
class PieStatus(Enum):
    """Status of the pie based on the set goal"""
    AHEAD = "AHEAD"
    ON_TRACK = "ON_TRACK"
    BEHIND = "BEHIND"

class DividendDetails(BaseModel):
    gained:Optional[float]
    inCash:Optional[float]
    reinvested:Optional[float]

class PieListItem(BaseModel):
    cash:Optional[float] # Amount of money put into the pie in account currency
    dividendDetails:DividendDetails
    id:Optional[int]
    progress:Optional[float] # Progress of the pie based on the set goal
    result:Optional[Result]
    status:Optional[PieStatus]

class PieList(BaseModel):
    pies:List[PieListItem]

'''
Equity Orders Classes
'''
class Order(BaseModel):
    creationTime: Optional[str] = None
    filledQuantity: Optional[float] = None
    filledValue: Optional[float] = None
    id: Optional[int] = None
    status: Optional[OrderStatus] = None
    strategy: Optional[str] = None
    type: Optional[str] = None
    value: Optional[float] = None
    # New Order Parameters
    limitPrice: Optional[float] = None
    quantity: Optional[float] = None
    ticker: Optional[str] = None
    timeValidity: Optional[str] = None
    stopPrice: Optional[str] = None

class Orders(BaseModel):
    orders: List[Order]

'''
Historical Items Classes
'''
class Tax(BaseModel):
    fillId: Optional[str]
    name: Optional[str]
    quantity: Optional[float]
    timeCharged: Optional[datetime]

class HistoricalOrder(BaseModel):
    dateCreated: Optional[datetime]
    dateExecuted: Optional[datetime]
    dateModified: Optional[datetime]
    executor: Optional[str]
    fillCost: Optional[float]
    fillId: Optional[int]
    fillPrice: Optional[float]
    fillResult: Optional[float]
    fillType: Optional[str]
    filledQuantity: Optional[float]
    filledValue: Optional[float]
    id: Optional[int]
    limitPrice: Optional[float]
    orderedQuantity: Optional[float]
    orderedValue: Optional[float]
    parentOrder: Optional[int]
    status: Optional[str]
    stopPrice: Optional[int]
    taxes: Optional[List[Tax]]
    ticker: Optional[str]
    timeValidity: Optional[str]
    type: Optional[str]

class HistoricalOrderResponseModel(BaseModel):
    items: Optional[List[HistoricalOrder]]
    nextPagePath: Optional[str] = None

class HistoricalItem(BaseModel):
    cursor:Optional[int]=None
    ticker:Optional[str]=None
    limit:Optional[int]=None

# Dividend
class DividendItem(BaseModel):
    amount: Optional[float]
    amountInEuro: Optional[float]
    grossAmountPerShare: Optional[float]
    paidOn: Optional[datetime]
    quantity: Optional[float]
    reference: Optional[str]
    ticker: Optional[str]
    type: Optional[str]

class DividendResponseModel(BaseModel):
    items: Optional[List[DividendItem]]
    nextPagePath: Optional[str]

# Export List
class DataIncluded(BaseModel):
    includeDividends: bool
    includeInterest: bool
    includeOrders: bool
    includeTransactions: bool

class ExportReport(BaseModel):
    dataIncluded: Optional[DataIncluded]
    downloadLink: Optional[str]
    reportId: Optional[int]
    status: Optional[str]
    timeFrom: Optional[datetime]
    timeTo: Optional[datetime]
    response: Optional[str]

# Exports
class DataIncluded(BaseModel):
    includeDividends: bool
    includeInterest: bool
    includeOrders: bool
    includeTransactions: bool

class ExportPayload(BaseModel):
    dataIncluded: DataIncluded
    timeFrom: datetime
    timeTo: datetime

class ExportReportResponse(BaseModel):
    reportId: Optional[int]

# Transaction List
class TransactionPayload(BaseModel):
    cursor: str
    limit: int = 20

class TransactionItem(BaseModel):
    amount: Optional[float]
    dateTime: Optional[datetime]
    reference: Optional[str]
    type: Optional[TransactionType]

class Transactions(BaseModel):
    items: Optional[List[TransactionItem]]
    nextPagePath: Optional[str]